"""
Tests for the decorator that converts a function to a command.
"""

from shellish import autocommand
import unittest


class PositionalTests(unittest.TestCase):

    def test_one_pos(self):
        @autocommand
        def f(one):
            self.assertEqual(one, 'ONE')
        f(argv='ONE')
        with self.assertRaises(SystemExit):
            f(argv='')

    def test_2_and_3_pos(self):
        @autocommand
        def f2(one, two):
            self.assertEqual(one, 'ONE')
            self.assertEqual(two, 'TWO')
        f2(argv='ONE TWO')
        with self.assertRaises(SystemExit):
            f2(argv='ONE')

        @autocommand
        def f3(one, two, three):
            self.assertEqual(one, 'ONE')
            self.assertEqual(two, 'TWO')
            self.assertEqual(three, 'THREE')
        f3(argv='ONE TWO THREE')
        with self.assertRaises(SystemExit):
            f3(argv='ONE TWO')

    def test_only_varargs(self):
        @autocommand
        def f(*args):
            self.assertEqual(args[0], 'ONE')
            self.assertEqual(args[1], 'TWO')
            self.assertEqual(len(args), 2)
        f(argv='ONE TWO')

    def test_one_pos_and_varargs(self):
        @autocommand
        def f(one, *args):
            self.assertEqual(one, 'posONE')
            self.assertEqual(args[0], 'ONE')
            self.assertEqual(args[1], 'TWO')
            self.assertEqual(len(args), 2)
        f(argv='posONE ONE TWO')

    def test_2_pos_and_varargs(self):
        @autocommand
        def f(one, two, *args):
            self.assertEqual(one, 'posONE')
            self.assertEqual(two, 'posTWO')
            self.assertEqual(args[0], 'ONE')
            self.assertEqual(args[1], 'TWO')
            self.assertEqual(len(args), 2)
        f(argv='posONE posTWO ONE TWO')

    def test_empty_varargs(self):
        @autocommand
        def f(*args):
            self.assertEqual(len(args), 0)
        f(argv='')


class KeywordTests(unittest.TestCase):

    def test_one_keyword(self):
        @autocommand
        def f(one=None):
            self.assertEqual(one, 'ONE')
        f(argv='--one ONE')

    def test_two_keywords(self):
        @autocommand
        def f(one=None, two=None):
            self.assertEqual(one, 'ONE')
            self.assertEqual(two, 'TWO')
        f(argv='--one ONE --two TWO')

    def test_only_varkwargs(self):
        @autocommand
        def f(**kwargs):
            self.assertEqual(kwargs['one'], 'ONE')
            self.assertEqual(kwargs['two'], 'TWO')
            self.assertEqual(len(kwargs), 2)
        f(argv='--kwargs --one ONE --two TWO')

    def test_onekw_and_varkwargs(self):
        @autocommand
        def f(first=None, **kwargs):
            self.assertEqual(first, 'FIRST')
            self.assertEqual(kwargs['one'], 'ONE')
            self.assertEqual(kwargs['two'], 'TWO')
            self.assertEqual(len(kwargs), 2)
        f(argv='--first FIRST --kwargs --one ONE --two TWO')

    def test_varkwargs_mixed_patterns(self):
        @autocommand
        def f(**kwargs):
            self.assertEqual(kwargs['one'], 'ONE')
            self.assertEqual(kwargs['two'], 'TWO')
            self.assertEqual(kwargs['three'], 'THREE')
            self.assertEqual(len(kwargs), 3)
        f(argv='--kwargs --one ONE --two TWO --three THREE')
        f(argv='--kwargs one=ONE two=TWO three=THREE')
        f(argv='--kwargs --one ONE two=TWO --three THREE')
        f(argv='--kwargs one=ONE --two TWO three=THREE')

    def test_empty_varkwargs(self):
        @autocommand
        def f(**kwargs):
            self.assertEqual(len(kwargs), 0)
        f(argv='')


class CombinationTests(unittest.TestCase):

    def test_arg_kwarg_varargs(self):
        with self.assertRaisesRegex(ValueError, 'Unsupported'):
            @autocommand
            def f(one, first=None, *args):
                pass

    def test_arg_varargs_kwarg(self):
        @autocommand
        def f(one, *args, first=None):
            self.assertEqual(one, 'ONE')
            self.assertEqual(args[0], 'VONE')
            self.assertEqual(args[1], 'VTWO')
            self.assertEqual(len(args), 2)
            self.assertEqual(first, 'FIRST')
        f(argv='ONE VONE VTWO --first FIRST')

    def test_arg_varkwarg(self):
        @autocommand
        def f(one, **kwargs):
            self.assertEqual(one, 'ONE')
            self.assertEqual(kwargs['kwone'], 'KWONE')
            self.assertEqual(kwargs['kwtwo'], 'KWTWO')
            self.assertEqual(len(kwargs), 2)
        f(argv='ONE --kwargs --kwone KWONE kwtwo=KWTWO')

    def test_arg_kwarg_defaultfailback(self):
        @autocommand
        def f(one, first='DEFAULT'):
            self.assertEqual(one, 'ONE')
            self.assertEqual(first, 'DEFAULT')
        f(argv='ONE')

    def test_arg_kwarg(self):
        @autocommand
        def f(one, first='DEFAULT'):
            self.assertEqual(one, 'ONE')
            self.assertEqual(first, 'FIRST')
        f(argv='ONE --first FIRST')

    def test_arg_varargs_kwarg_varkwargs(self):
        @autocommand
        def f(one, *args, first='DEFAULT', **kwargs):
            self.assertEqual(one, 'ONE')
            self.assertEqual(args[0], 'VONE')
            self.assertEqual(args[1], 'VTWO')
            self.assertEqual(first, 'FIRST')
            self.assertEqual(kwargs['kwone'], 'KWONE')
            self.assertEqual(kwargs['kwtwo'], 'KWTWO')
            self.assertEqual(len(kwargs), 2)
        f(argv='ONE VONE VTWO --first FIRST --kwargs kwone=KWONE kwtwo=KWTWO')

    def test_empty_var_and_keyword_args(self):
        @autocommand
        def f(*args, **kwargs):
            self.assertEqual(len(args), 0)
            self.assertEqual(len(kwargs), 0)
        f(argv='')


class TypeTests(unittest.TestCase):
    """ Tests where the arguments gather type info from the function
    signature. """

    def test_annotation_str(self):
        @autocommand
        def f(one: str):
            self.assertIsInstance(one, str)
        f(argv='asdf')

    def test_pos_annotation_bool(self):
        @autocommand
        def f(one: bool):
            self.assertIsInstance(one, bool)
            return one
        for x in ('False', 'no', 'disable', '0', 'null', 'None'):
            with self.subTest(x=x):
                self.assertFalse(f(argv=x))
        for x in ('1', 'yes', 'True', 'randomthing'):
            with self.subTest(x=x):
                self.assertTrue(f(argv=x))

    def test_kw_bool_toggle(self):
        """ Bool keywords operate like toggles.  When present on the arg line, they
        invert the default. """
        @autocommand
        def f(toggle=False):
            self.assertIsInstance(toggle, bool)
            return toggle
        self.assertFalse(f(argv=''))
        self.assertTrue(f(argv='--toggle'))

        @autocommand
        def f(toggle=True):
            self.assertIsInstance(toggle, bool)
            return toggle
        self.assertTrue(f(argv=''))
        self.assertFalse(f(argv='--toggle'))

    def test_annotation_int(self):
        @autocommand
        def f(one: int):
            self.assertIsInstance(one, int)
            return one
        for x in [0, 1, 2**1024, -1, -2]:
            self.assertEqual(f(argv=str(x)), x)
        for x in ['nope', '0x0', '0o0']:
            with self.assertRaises(SystemExit):
                f(argv=x)

    def test_annotation_float(self):
        @autocommand
        def f(one: float):
            self.assertIsInstance(one, float)
            return one
        for x in [0, 1, 1.1]:
            self.assertEqual(f(argv=str(x)), x)
        for x in ['nope']:
            with self.assertRaises(SystemExit):
                f(argv=x)


class Nesting(unittest.TestCase):

    def test_one_level(self):
        @autocommand
        def main():
            return 'main'

        @autocommand
        def sub():
            return 'sub'

        main.add_subcommand(sub)
        self.assertEqual(main(argv=''), 'main')
        self.assertEqual(main(argv='sub'), 'sub')



"""
Sanity tests for the shellish library.
"""

import shellish
import unittest


class CommandSanity(unittest.TestCase):

    def test_command_init(self):
        class Foo(shellish.Command):
            """ foo """
            pass
        Foo(name='foo')
        shellish.Command(name='test')

    def test_no_title_desc_by_subclass(self):
        class Foo(shellish.Command):
            pass
        f = Foo(name='foo')
        self.assertIsNone(f.title)
        self.assertIsNone(f.desc)

    def test_no_title_desc_by_compose(self):
        f = shellish.Command(name='foo')
        self.assertIsNone(f.title)
        self.assertIsNone(f.desc)

    def test_title_desc_by_subclass_attrs(self):
        class Foo(shellish.Command):
            title = 'foo'
            desc = 'bar'
        f = Foo(name='foo')
        self.assertEqual(f.title, 'foo')
        self.assertEqual(f.desc, 'bar')

    def test_title_desc_by_subclass_docstring_2line_padded(self):
        class Foo(shellish.Command):
            """ foo 
            bar """  # noqa
        f = Foo(name='foo')
        self.assertEqual(f.title, 'foo')
        self.assertEqual(f.desc, 'bar')

    def test_title_desc_by_subclass_docstring_2line_unpadded(self):
        class Foo(shellish.Command):
            """foo
            bar"""
        f = Foo(name='foo')
        self.assertEqual(f.title, 'foo')
        self.assertEqual(f.desc, 'bar')

    def test_title_desc_by_subclass_docstring_3line_padded_gap(self):
        class Foo(shellish.Command):
            """ foo 

            bar """  # noqa
        f = Foo(name='foo')
        self.assertEqual(f.title, 'foo')
        self.assertEqual(f.desc, 'bar')

    def test_title_desc_by_subclass_docstring_3line_extrahead(self):
        class Foo(shellish.Command):
            """
            foo
            bar """
        f = Foo(name='foo')
        self.assertEqual(f.title, 'foo')
        self.assertEqual(f.desc, 'bar')

    def test_title_desc_by_subclass_docstring_3line_extrafoot(self):
        class Foo(shellish.Command):
            """foo
            bar
            """
        f = Foo(name='foo')
        self.assertEqual(f.title, 'foo')
        self.assertEqual(f.desc, 'bar')

    def test_title_desc_by_subclass_docstring_4line(self):
        class Foo(shellish.Command):
            """
            foo
            bar
            """
        f = Foo(name='foo')
        self.assertEqual(f.title, 'foo')
        self.assertEqual(f.desc, 'bar')

    def test_empty_docstring(self):
        class Foo(shellish.Command):
            """"""
        f = Foo(name='foo')
        self.assertEqual(f.title, None)
        self.assertEqual(f.desc, None)

        class Foo(shellish.Command):
            """ """
        f = Foo(name='foo')
        self.assertEqual(f.title, None)
        self.assertEqual(f.desc, None)

        class Foo(shellish.Command):
            """
            """
        f = Foo(name='foo')
        self.assertEqual(f.title, None)
        self.assertEqual(f.desc, None)

        class Foo(shellish.Command):
            """

            """
        f = Foo(name='foo')
        self.assertEqual(f.title, None)
        self.assertEqual(f.desc, None)



import argparse
import shellish
import sys

mdfile = argparse.FileType('r', encoding='utf-8-sig')


@shellish.autocommand
def mdcat(mdfile: mdfile=sys.stdin):
    """ Pretty print a markdown file. """
    shellish.mdprint(mdfile.read())



"""
Command tree.
"""

import collections
from .. import command
from ... import layout


class Commands(command.Command):
    """ Show a command tree. """

    name = 'commands'
    use_pager = True

    def setup_args(self, parser):
        pass

    def command_choices(self, prefix, args):
        return frozenset(x for x in self.parent.subcommands
                         if x.startswith(prefix))

    def run(self, args):
        root = self.parent
        tree = self.walkinto(root)
        layout.treeprint({root.name: tree})

    def walkinto(self, command):
        tree = collections.OrderedDict()
        if not command.subcommands:
            return command.title or command.name
        else:
            for key, cmd in command.subcommands.items():
                tree[key] = self.walkinto(cmd)
        return tree




import os
import shellish
import sys
import tempfile
import unittest


class TestNesting(unittest.TestCase):

    def test_nest1_by_class(self):

        class A(shellish.Command):
            name = 'a'

            def run(self, args):
                return 'A'

        class B(shellish.Command):
            name = 'b'

            def run(self, args):
                return 'B'

        class B2(shellish.Command):
            name = 'b2'

            def run(self, args):
                return 'B2'
        a = A()
        a.add_subcommand(B)
        a.add_subcommand(B2())
        self.assertEqual(a(argv=''), 'A')
        self.assertEqual(a(argv='b'), 'B')
        self.assertEqual(a(argv='b2'), 'B2')

    def test_default_subcommand(self):

        class A(shellish.Command):
            name = 'a'

            def run(self, args):
                return 'A'

        class B(shellish.Command):
            name = 'b'

            def run(self, args):
                return 'B'
        a = A()
        a.add_subcommand(B, default=True)
        self.assertEqual(a(argv=''), 'B')
        self.assertEqual(a(argv='b'), 'B')


class CommandCompose(unittest.TestCase):

    def test_naked(self):
        a = shellish.Command(name='a')
        self.assertRaises(SystemExit, a, argv='')

    def test_onlyrun(self):
        ok = object()
        a = shellish.Command(name='a', run=lambda _: ok)
        self.assertIs(a(argv=''), ok)

    def test_allrunners(self):
        refcnt = 0

        def runner(args, **ign):
            nonlocal refcnt
            refcnt += 1
        a = shellish.Command(name='a', prerun=runner, run=runner,
                             postrun=runner)
        a(argv='')
        self.assertEqual(refcnt, 3)

    def test_command_nameless(self):
        self.assertRaises(RuntimeError, shellish.Command)

    def test_subcommand_name_at_construct(self):
        a = shellish.Command(name='a')
        b = shellish.Command(name='b')
        a.add_subcommand(b)


class CommandFileArguments(unittest.TestCase):

    def setUp(self):
        self._cwd_save = os.getcwd()
        self.tmp = tempfile.TemporaryDirectory()
        os.chdir(self.tmp.name)

    def tearDown(self):
        os.chdir(self._cwd_save)
        self.tmp.cleanup()

    def test_file_argument_defaults_not_found(self):
        def run(args):
            self.assertRaises(FileNotFoundError, args.foo().__enter__)
        cmd = shellish.Command(name='test', run=run)
        cmd.add_file_argument('--foo')
        cmd(argv='--foo doesnotexist')

    def test_file_argument_defaults_stdio(self):
        flushed = False

        class Stdin(object):

            def flush(self):
                nonlocal flushed
                flushed = True
        stdin_setinel = Stdin()

        def run(args):
            with args.foo as f:
                self.assertIs(f, stdin_setinel)
                self.assertFalse(flushed)
            self.assertTrue(flushed)
        cmd = shellish.Command(name='test', run=run)
        cmd.add_file_argument('--foo')
        stdin = sys.stdin
        sys.stdin = stdin_setinel
        try:
            cmd(argv='--foo -')
        finally:
            sys.stdin = stdin

    def test_file_argument_stdout(self):
        def run(args):
            with args.foo as f:
                self.assertIs(f, sys.stdout)
        cmd = shellish.Command(name='test', run=run)
        cmd.add_file_argument('--foo', mode='w')
        cmd(argv='--foo -')

    def test_file_argument_read_found(self):
        def run(args):
            with args.foo as f:
                self.assertTrue(f.name.endswith('exists'))
                self.assertTrue(f.mode, 'r')
        open('exists', 'w').close()
        cmd = shellish.Command(name='test', run=run)
        cmd.add_file_argument('--foo', mode='r')
        cmd(argv='--foo exists')
        cmd(argv='--foo ./exists')

    def test_file_argument_create(self):
        def run(args):
            with args.foo as f:
                self.assertEqual(f.name, 'makethis')
                f.write('ascii is okay')
        cmd = shellish.Command(name='test', run=run)
        cmd.add_file_argument('--foo', mode='w')
        cmd(argv='--foo makethis')


class ArgumentDestSafety(unittest.TestCase):

    def test_dash_positional(self):
        c = shellish.Command(name='test_dash')
        c.add_argument('foo-bar')
        args = c.argparser.parse_known_args(['value'])[0]
        self.assertIn('foo_bar', args)
        self.assertEqual(args.foo_bar, 'value')

    def test_dash_opt(self):
        c = shellish.Command(name='test_dash')
        c.add_argument('--foo-bar')
        args = c.argparser.parse_known_args(['--foo-bar', 'value'])[0]
        self.assertIn('foo_bar', args)
        self.assertEqual(args.foo_bar, 'value')



"""
Support for directing output to a pager and then restoring tty state.
"""

import contextlib
import fcntl
import os
import shutil
import signal
import subprocess
import sys
import termios
import warnings

_pager_active = False


@contextlib.contextmanager
def tty_restoration(file=None):
    if file is None:
        file = sys.stdout
    assert file.isatty()
    tcsave = termios.tcgetattr(file)
    fcsave = fcntl.fcntl(file, fcntl.F_GETFL)
    try:
        yield
    finally:
        fcntl.fcntl(file, fcntl.F_SETFL, fcsave)
        termios.tcsetattr(file, termios.TCSADRAIN, tcsave)


def pager_process(pagercmd, stdout=None, stderr=None):
    if stdout is None:
        stdout = sys.stdout
    if stderr is None:
        stderr = sys.stderr
    # When running interactively `less` does not handle window resizes
    # unless we explicitly hardcode the new term size into the env.  There
    # is currently a bug in docker for mac that sometimes breaks this test.
    env = os.environ.copy()
    termsize = shutil.get_terminal_size()
    if 0 in termsize:
        warnings.warn("Could not determine terminal size")
        termsize = os.terminal_size((80, 24))
    env['COLUMNS'] = str(termsize.columns)
    env['LINES'] = str(termsize.lines)
    return subprocess.Popen(pagercmd, shell=True, universal_newlines=True,
                            bufsize=1, stdin=subprocess.PIPE, stdout=stdout,
                            stderr=stderr, env=env)


@contextlib.contextmanager
def pager_redirect(desc, *, pagercmd=None, istty=None, file=None,
                   substitutions=None):
    """ Redirect output to file/stdout to a pager process.  Care is taken to
    restore the controlling tty stdio files to their original state. """
    global _pager_active
    if file is None:
        file = sys.stdout
    if not pagercmd or not file.isatty() or _pager_active:
        yield
        return
    subs = {"desc": desc}
    if substitutions is not None:
        subs.update(substitutions)
    pagercmd = pagercmd.format(**subs)
    with tty_restoration():
        p = pager_process(pagercmd)
        if istty is None:
            p.stdin.isatty = file.isatty
        else:
            p.stdin.isatty = lambda: istty
        stdout_save = sys.stdout
        sys.stdout = p.stdin
        _pager_active = True
        try:
            yield
        finally:
            _pager_active = False
            sys.stdout = stdout_save
            try:
                p.stdin.close()
            except BrokenPipeError:
                pass
            while p.poll() is None:
                try:
                    p.wait()
                except KeyboardInterrupt:
                    pass



"""
Nesting commands using simple syntax.
"""

import shellish


@shellish.autocommand
def main():
    print("Default Action")


@shellish.autocommand
def sub(option=None):
    print("Hi from sub", option)


main.add_subcommand(sub)
main()



"""
Show the INI config(s) used by a command tree.
"""

from .. import command


class Show(command.Command):
    """ Show current INI configuration.

    Programs may make use of a configuration file which is usually located in
    your $HOME directory as .<prog>_config.  The file is a standard INI
    style config file where each `[section]` is the full path of a command
    including spaces. """

    name = 'show'

    def setup_args(self, parser):
        self.add_argument('section', nargs='?', help='Only show config for '
                          'this section.')
        self.add_argument('--all', '-a', action='store_true', help='Show '
                          'all sections')
        super().setup_args(parser)

    def run(self, args):
        if args.section:
            try:
                config = {args.section: self.session.config[args.section]}
            except KeyError:
                raise SystemExit("Invalid section: %s" % args.section)
        else:
            config = self.session.config
        for section, values in config.items():
            if values or args.all:
                print("[%s]" % section)
                for k, v in values.items():
                    print("     %s = %s" % (k, v))
                print()


class INI(command.Command):
    """ INI style configuration.
    Commands support user configuration in an INI style config file. """

    name = 'ini'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_subcommand(Show, default=True)




import itertools
import unittest
from shellish import rendering as R


class VTMLBufferTests(unittest.TestCase):

    def test_overclip_plain(self):
        startval = 'A' * 10
        s = R.vtmlrender(startval)
        self.assertEqual(s.clip(11), startval)
        self.assertEqual(s.clip(11).text(), startval)
        self.assertEqual(s.clip(20), startval)
        self.assertEqual(s.clip(20).text(), startval)

    def test_noclip_plain(self):
        startval = 'A' * 10
        s = R.vtmlrender(startval)
        self.assertEqual(s.clip(10), startval)
        self.assertEqual(s.clip(10).text(), startval)

    def test_underclip_plain(self):
        startval = 'A' * 10
        s = R.vtmlrender(startval)
        self.assertEqual(s.clip(9), startval[:9])
        self.assertEqual(s.clip(9).text(), startval[:9])
        self.assertEqual(s.clip(4), startval[:4])
        self.assertEqual(s.clip(4).text(), startval[:4])
        self.assertEqual(s.clip(1), startval[:1])
        self.assertEqual(s.clip(1).text(), startval[:1])
        self.assertEqual(s.clip(0), '')
        self.assertEqual(s.clip(0).text(), '')
        self.assertRaises(ValueError, s.clip, -10)

    def test_overclip_vtml(self):
        startval = 'A' * 10
        s = R.vtmlrender('<b>%s</b>' % startval)
        self.assertEqual(s.clip(11).text(), startval)
        self.assertEqual(s.clip(20).text(), startval)
        self.assertEqual(s.clip(11), s)
        self.assertEqual(s.clip(20), s)

    def test_noclip_vtml(self):
        startval = 'A' * 10
        s = R.vtmlrender('<b>%s</b>' % startval)
        self.assertEqual(s.clip(10).text(), startval)
        self.assertEqual(s.clip(10), s)

    def test_underclip_vtml(self):
        startval = 'A' * 10
        s = R.vtmlrender('<b>%s</b>' % startval)
        self.assertEqual(s.clip(9).text(), startval[:9])
        self.assertEqual(str(s.clip(9)).count('A'), 9)
        self.assertEqual(s.clip(4).text(), startval[:4])
        self.assertEqual(str(s.clip(4)).count('A'), 4)
        self.assertEqual(s.clip(1).text(), startval[:1])
        self.assertEqual(str(s.clip(1)).count('A'), 1)
        self.assertEqual(s.clip(0).text(), '')
        self.assertEqual(s.clip(0), '')
        self.assertRaises(ValueError, s.clip, -10)

    def test_underclip_vtml_reset(self):
        s = R.vtmlrender('<b>%s</b>' % 'AAAA')
        self.assertTrue(str(s.clip(2)).endswith('\033[0m'))

    def test_overclip_with_cliptextt(self):
        startval = 'A' * 10
        s = R.vtmlrender(startval)
        self.assertEqual(s.clip(12, '.'), startval)
        self.assertEqual(s.clip(11, '.'), startval)
        self.assertEqual(s.clip(10, '.'), startval)
        self.assertEqual(s.clip(12, '..'), startval)
        self.assertEqual(s.clip(11, '..'), startval)
        self.assertEqual(s.clip(10, '..'), startval)
        self.assertEqual(s.clip(12, '...'), startval)
        self.assertEqual(s.clip(11, '...'), startval)
        self.assertEqual(s.clip(10, '...'), startval)

    def test_underclip_with_cliptextt(self):
        startval = 'A' * 10
        s = R.vtmlrender(startval)
        self.assertEqual(s.clip(9, '.'), startval[:8] + '.')
        self.assertEqual(s.clip(8, '.'), startval[:7] + '.')
        self.assertEqual(s.clip(7, '.'), startval[:6] + '.')
        self.assertEqual(s.clip(9, '..'), startval[:7] + '..')
        self.assertEqual(s.clip(8, '..'), startval[:6] + '..')
        self.assertEqual(s.clip(7, '..'), startval[:5] + '..')
        self.assertEqual(s.clip(9, '...'), startval[:6] + '...')
        self.assertEqual(s.clip(8, '...'), startval[:5] + '...')
        self.assertEqual(s.clip(7, '...'), startval[:4] + '...')
        self.assertEqual(s.clip(6, '...'), startval[:3] + '...')

    def test_clip_strip(self):
        self.assertEqual(R.vtmlrender('\n').clip(10, '!').text(), '!')
        self.assertEqual(R.vtmlrender(' \n').clip(10, '!').text(), '!')
        self.assertEqual(R.vtmlrender(' ').clip(10, '!').text(), '')
        self.assertEqual(R.vtmlrender('A\n').clip(10, '!').text(), 'A!')
        self.assertEqual(R.vtmlrender('A\nB').clip(10, '!').text(), 'A!')
        self.assertEqual(R.vtmlrender('A\n').clip(1, '!').text(), '!')
        self.assertEqual(R.vtmlrender('A ').clip(1, '!').text(), 'A')
        self.assertEqual(R.vtmlrender('A  ').clip(1, '!').text(), 'A')
        self.assertEqual(R.vtmlrender('A \n').clip(1, '!').text(), '!')
        self.assertEqual(R.vtmlrender('A \n ').clip(1, '!').text(), '!')
        self.assertEqual(R.vtmlrender('A\n').clip(2, '!').text(), 'A!')
        self.assertEqual(R.vtmlrender('A ').clip(2, '!').text(), 'A')
        self.assertEqual(R.vtmlrender('A  ').clip(2, '!').text(), 'A')
        self.assertEqual(R.vtmlrender('A \n').clip(2, '!').text(), 'A!')
        self.assertEqual(R.vtmlrender('A \n ').clip(2, '!').text(), 'A!')
        self.assertEqual(R.vtmlrender('    ').clip(5, '!').text(), '')
        self.assertEqual(R.vtmlrender('    ').clip(4, '!').text(), '')
        self.assertEqual(R.vtmlrender('    ').clip(3, '!').text(), '')
        self.assertEqual(R.vtmlrender('    ').clip(2, '!').text(), '')
        self.assertEqual(R.vtmlrender('    ').clip(1, '!').text(), '')

    def test_clip_empty(self):
        self.assertEqual(R.vtmlrender('').clip(10, '!').text(), '')

    def test_wrap_empty(self):
        buf = R.vtmlrender('')
        self.assertListEqual(buf.wrap(10), [''])

    def test_wrap_identity(self):
        buf = R.vtmlrender('abcdefgh')
        self.assertIsInstance(buf.wrap(10), list)
        self.assertIsInstance(buf.wrap(8), list)
        self.assertIsInstance(buf.wrap(1), list)
        self.assertRaises(ValueError, buf.wrap, -1)
        self.assertRaises(ValueError, buf.wrap, 0)
        self.assertRaises(ValueError, buf.wrap, -2)
        self.assertRaises(ValueError, buf.wrap, -8)
        self.assertRaises(ValueError, buf.wrap, -10)

    def test_wrap_boundries_packed_stronly(self):
        buf = R.vtmlrender('abcdefgh')
        self.assertListEqual(buf.wrap(10), ['abcdefgh'])
        self.assertListEqual(buf.wrap(8), ['abcdefgh'])
        self.assertListEqual(buf.wrap(7), ['abcdefg', 'h'])
        self.assertListEqual(buf.wrap(2), ['ab', 'cd', 'ef', 'gh'])
        self.assertListEqual(buf.wrap(1), list('abcdefgh'))

    def test_wrap_boundries_hypens_stronly(self):
        buf = R.vtmlrender('abcd-efgh')
        self.assertListEqual(buf.wrap(10), ['abcd-efgh'])
        self.assertListEqual(buf.wrap(8), ['abcd-', 'efgh'])

    def test_wrap_whitespace_stronly(self):
        buf = R.vtmlrender('abcd efgh')
        self.assertListEqual(buf.wrap(10), ['abcd efgh'])
        self.assertListEqual(buf.wrap(8), ['abcd', 'efgh'])
        self.assertListEqual(buf.wrap(7), ['abcd', 'efgh'])
        self.assertListEqual(buf.wrap(2), ['ab', 'cd', 'ef', 'gh'])
        self.assertListEqual(buf.wrap(1), list('abcdefgh'))

    def test_wrap_multi_whitespace_stronly(self):
        buf = R.vtmlrender(' A BB  CCC  DDDD EEEEE ')
        self.assertListEqual(buf.wrap(10), [' A BB  CCC', 'DDDD EEEEE'])
        self.assertListEqual(buf.wrap(8), [' A BB', 'CCC', 'DDDD', 'EEEEE'])
        self.assertListEqual(buf.wrap(7), [' A BB', 'CCC', 'DDDD', 'EEEEE'])
        self.assertListEqual(buf.wrap(2), [
            ' A', 'BB', 'CC', 'C', 'DD', 'DD', 'EE', 'EE', 'E'])
        self.assertListEqual(buf.wrap(1), [
            ' ', 'A', 'B', 'B', 'C', 'C', 'C', 'D', 'D', 'D', 'D', 'E', 'E',
            'E', 'E', 'E'])

    def test_wrap_overflow_whitespace_stronly(self):
        buf = R.vtmlrender('        A BB  ')
        self.assertListEqual(buf.wrap(20), ['        A BB'])
        self.assertListEqual(buf.wrap(5), ['     ', 'A BB'])
        self.assertListEqual(buf.wrap(3), ['   ', 'A', 'BB'])
        self.assertListEqual(buf.wrap(2), ['  ', 'A', 'BB'])
        buf = R.vtmlrender('        A BB         ')
        self.assertListEqual(buf.wrap(20), ['        A BB'])
        self.assertListEqual(buf.wrap(5), ['     ', 'A BB'])
        self.assertListEqual(buf.wrap(3), ['   ', 'A', 'BB'])
        self.assertListEqual(buf.wrap(2), ['  ', 'A', 'BB'])
        buf = R.vtmlrender('        A             BB         ')
        self.assertListEqual(buf.wrap(30), ['        A             BB'])
        self.assertListEqual(buf.wrap(5), ['     ', 'A', 'BB'])
        self.assertListEqual(buf.wrap(3), ['   ', 'A', 'BB'])
        self.assertListEqual(buf.wrap(2), ['  ', 'A', 'BB'])

    def test_wrap_newlines(self):
        for width in 1, 2, 80:
            with self.subTest('width=%d' % width):
                buf = R.vtmlrender('A\nB')
                self.assertListEqual(buf.wrap(width), ['A', 'B'])
                buf = R.vtmlrender('A\n\nB')
                self.assertListEqual(buf.wrap(width), ['A', '', 'B'])
                buf = R.vtmlrender('A\n\n\nB')
                self.assertListEqual(buf.wrap(width), ['A', '', '', 'B'])
                buf = R.vtmlrender('A\n\n\nB')
                self.assertListEqual(buf.wrap(width), ['A', '', '', 'B'])
                buf = R.vtmlrender('A\n   B')
                self.assertListEqual(buf.wrap(width), ['A', 'B'])
                buf = R.vtmlrender('A    \n   B')
                self.assertListEqual(buf.wrap(width), ['A', 'B'])
                buf = R.vtmlrender('A    \nB')
                self.assertListEqual(buf.wrap(width), ['A', 'B'])

    def test_wrap_no_indent(self):
        buf = R.vtmlrender('  A\nB')
        self.assertListEqual(buf.wrap(20, strip_leading_indent=True),
                             ['A', 'B'])

    @unittest.skip
    def test_wrap_expand_tabs(self):
        buf = R.vtmlrender('A\tB')
        self.assertListEqual(buf.wrap(20), ['A    B'])
        self.assertListEqual(buf.wrap(3), ['A', 'B'])
        self.assertListEqual(buf.wrap(2), ['A', 'B'])
        self.assertListEqual(buf.wrap(20, expand_tabs=False), ['A\tB'])
        self.assertListEqual(buf.wrap(3, expand_tabs=False), ['A\tB'])
        self.assertListEqual(buf.wrap(2, expand_tabs=False), ['A', 'B'])

    def test_bad_data(self):
        bad = [
            None,
            0,
            1,
            ['asdf', 'asdf'],
            [None, None],
            '<>asdf',
            '</b>asdf</notit>',
        ]
        for x in bad:
            self.assertEqual(R.vtmlrender(x), x)

    def test_ordering(self):
        self.assertGreater(R.vtmlrender('bbbb'), R.vtmlrender('aaaa'))
        self.assertLess(R.vtmlrender('aaaa'), R.vtmlrender('bbbb'))
        self.assertGreaterEqual(R.vtmlrender('bbbb'), R.vtmlrender('aaaa'))
        self.assertGreaterEqual(R.vtmlrender('aaaa'), R.vtmlrender('aaaa'))
        self.assertLessEqual(R.vtmlrender('aaaa'), R.vtmlrender('bbbb'))
        self.assertLessEqual(R.vtmlrender('aaaa'), R.vtmlrender('aaaa'))
        self.assertEqual(R.vtmlrender('aaaa'), R.vtmlrender('aaaa'))

    def test_add_same_type(self):
        a = R.vtmlrender('aaaa')
        b = R.vtmlrender('BBBB')
        c = a + b
        self.assertIsNot(c, a)
        self.assertIsNot(c, b)
        self.assertEqual(c, R.vtmlrender('aaaaBBBB'))
        self.assertEqual(str(c), 'aaaaBBBB')

    def test_add_str_type(self):
        a = R.vtmlrender('aaaa')
        b = 'BBBB'
        c = a + b
        self.assertIsNot(c, a)
        self.assertEqual(c, R.vtmlrender('aaaaBBBB'))
        self.assertEqual(str(c), 'aaaaBBBB')

    def test_iadd_same_type(self):
        a1 = a1_save = R.vtmlrender('aaaa')
        a1 += R.vtmlrender('BBBB')
        self.assertIs(a1, a1_save)
        self.assertEqual(a1, R.vtmlrender('aaaaBBBB'))
        self.assertEqual(str(a1), 'aaaaBBBB')

    def test_iadd_str_type(self):
        a1 = a1_save = R.vtmlrender('aaaa')
        a1 += 'BBBB'
        self.assertIs(a1, a1_save)
        self.assertEqual(a1, R.vtmlrender('aaaaBBBB'))
        self.assertEqual(str(a1), 'aaaaBBBB')

    def test_right_add_str_type(self):
        a = R.vtmlrender('aaaa')
        b = 'BBBB' + a
        self.assertEqual(b, 'BBBBaaaa')
        self.assertIsInstance(b, str)

    def test_add_ident(self):
        a1 = R.vtmlrender('aaaa')
        a2 = a1 + 'BBBB'
        self.assertIsNot(a1, a2)
        self.assertNotEqual(a1, a2)

    def test_iadd_unsupport_type(self):
        a1 = R.vtmlrender('foo')
        self.assertRaises(TypeError, lambda: a1 + 1)
        self.assertRaises(TypeError, lambda: a1 + b'bar')

    def test_amp_tail_single_char(self):
        """ Without a workaround this hits a bug in HTMLParser. """
        t = 'a&b'
        self.assertEqual(R.vtmlrender(t, strict=True), t)

    def test_amp_tail_double_char(self):
        t = 'a&bc'
        self.assertEqual(R.vtmlrender(t, strict=True), t)

    def test_amp_tail(self):
        t = 'a&'
        self.assertEqual(R.vtmlrender(t, strict=True), t)

    def test_amp_normal(self):
        for t in ('a&gt;', '&lt;', '&ltss;', '&;', '&abc;<other>'):
            self.assertEqual(R.vtmlrender(t, strict=True), t)

    def test_wrong_tag_tolerance(self):
        bad = ('foobar', 'foo', 'bar')
        perms = itertools.permutations(bad)
        for starts in perms:
            suffix = '<b>valid</b>'
            valid = str(R.vtmlrender(suffix, strict=True))
            self.assertIn('\033', valid, 'test to validate next portion')
            ends = next(perms)  # makes for bad closes
            buf = ["<%s>asdf</%s>" % (x, y) for x, y in zip(starts, ends)]
            line = ''.join(buf)
            ugly = str(R.vtmlrender(line + suffix, strict=True))
            self.assertIn(valid, ugly)
            self.assertEqual(ugly, line + valid, 'partial conv did not work')

    def test_pound(self):
        for t in ('a#bc', 'a&#1;'):
            self.assertEqual(R.vtmlrender(t, strict=True), t)

    def test_multiply(self):
        a = R.vtmlrender('A')
        self.assertEqual(a * 2, R.vtmlrender('AA'))
        self.assertEqual(a * 3, R.vtmlrender('AAA'))
        self.assertEqual((a * 3).text(), 'AAA')
        a = R.vtmlrender('AB')
        self.assertEqual(a * 2, R.vtmlrender('ABAB'))
        self.assertEqual(a * 3, R.vtmlrender('ABABAB'))
        self.assertEqual((a * 3).text(), 'ABABAB')
        self.assertRaises(TypeError, lambda: 1.5 * a)
        self.assertRaises(TypeError, lambda: '' * a)
        self.assertRaises(TypeError, lambda: a * a)

    def test_inplace_multiply(self):
        a = a_save = R.vtmlrender('A ')
        a *= 3
        self.assertIs(a, a_save)
        self.assertEqual(a, R.vtmlrender('A A A '))
        self.assertRaises(TypeError, lambda: 1.5 * a)
        self.assertRaises(TypeError, lambda: '' * a)
        self.assertRaises(TypeError, lambda: a * a)

    def test_right_multiply(self):
        a = R.vtmlrender('A')
        self.assertEqual(2 * a, R.vtmlrender('AA'))
        self.assertEqual(3 * a, R.vtmlrender('AAA'))
        self.assertEqual((3 * a).text(), 'AAA')
        a = R.vtmlrender('AB')
        self.assertEqual(2 * a, R.vtmlrender('ABAB'))
        self.assertEqual(3 * a, R.vtmlrender('ABABAB'))
        self.assertEqual((3 * a).text(), 'ABABAB')
        self.assertRaises(TypeError, lambda: 1.5 * a)
        self.assertRaises(TypeError, lambda: '' * a)
        self.assertRaises(TypeError, lambda: a * a)

    def test_left_justify(self):
        a = R.vtmlrender('A')
        self.assertEqual(a.ljust(0), 'A')
        self.assertEqual(a.ljust(1), 'A')
        self.assertEqual(a.ljust(2), 'A ')
        self.assertEqual(a.ljust(3), 'A  ')

    def test_startswith(self):
        abc = R.vtmlrender('abc')
        self.assertIs(abc.startswith(''), True)
        self.assertIs(abc.startswith('a'), True)
        self.assertIs(abc.startswith('ab'), True)
        self.assertIs(abc.startswith('abc'), True)
        self.assertIs(abc.startswith('A'), False)
        self.assertIs(abc.startswith('ABC'), False)
        self.assertIs(abc.startswith('ABCD'), False)
        a = R.vtmlrender('<b>a')
        b = R.vtmlrender('b')
        c = R.vtmlrender('<u>c')
        self.assertIs((a + b + c).startswith('abc'), True)
        self.assertIs((a + b + c).startswith(''), True)
        self.assertIs((a + b + c).startswith('<b>'), False)
        self.assertIs((a + b + c).startswith('A'), False)
        self.assertIs((a + b + c).startswith('ABC'), False)
        self.assertIs((a + b + c).startswith('ABCD'), False)

    def test_endswith(self):
        abc = R.vtmlrender('abc')
        self.assertIs(abc.endswith(''), True)
        self.assertIs(abc.endswith('c'), True)
        self.assertIs(abc.endswith('bc'), True)
        self.assertIs(abc.endswith('abc'), True)
        self.assertIs(abc.endswith('C'), False)
        self.assertIs(abc.endswith('ABC'), False)
        self.assertIs(abc.endswith('ABCD'), False)
        abc = R.vtmlrender('<b>a')
        abc += R.vtmlrender('b')
        abc += R.vtmlrender('<u>c</u>')
        self.assertIs(abc.endswith('abc'), True)
        self.assertIs(abc.endswith(''), True)
        self.assertIs(abc.endswith('</u>'), False)
        self.assertIs(abc.endswith('C'), False)
        self.assertIs(abc.endswith('ABC'), False)
        self.assertIs(abc.endswith('ABCD'), False)

    def test_in(self):
        abc = R.vtmlrender('abc')
        self.assertIn('', abc)
        self.assertIn('a', abc)
        self.assertIn('abc', abc)
        self.assertIn('abc', abc)
        self.assertNotIn('A', abc)
        self.assertNotIn('ABC', abc)
        self.assertNotIn('ABCD', abc)
        abc = R.vtmlrender('<b>a') + R.vtmlrender('bc') + R.vtmlrender('<u>c')
        self.assertIn('a', abc)
        self.assertIn('abc', abc)
        self.assertNotIn('<b>', abc)
        self.assertNotIn('A', abc)
        self.assertNotIn('ABC', abc)
        self.assertNotIn('ABCD', abc)
        self.assertNotIn('abcd', abc)

    def test_split_maxsplit(self):
        ref = 'abc ABC xyz XYZ'
        vs = R.vtmlrender(ref)
        self.assertListEqual(vs.split(), ref.split(' '))
        for i in range(6):
            with self.subTest(i):
                self.assertListEqual(vs.split(maxsplit=i), ref.split(' ', i))

    def test_split_leading(self):
        for ref in (' ', ' abc', '  ', '  abc'):
            vs = R.vtmlrender(ref)
            self.assertListEqual(vs.split(), ref.split(' '))

    def test_split_trailing(self):
        for ref in ('abc ', 'abc  ', ' abc ', '  abc  '):
            vs = R.vtmlrender(ref)
            self.assertListEqual(vs.split(), ref.split(' '))

    def test_split_notfound(self):
        ref = 'abc'
        vs = R.vtmlrender(ref)
        self.assertListEqual(vs.split('D'), ref.split('D'))
        self.assertListEqual(vs.split('abcd'), ref.split('abcd'))
        self.assertListEqual(vs.split('bcd'), ref.split('bcd'))

    def test_split_multichar(self):
        for ref in ('abcGAPABC', 'abcGAP', 'GAPabc', 'abGAPABGAP',
                    'aGAPbGAPc'):
            vs = R.vtmlrender(ref)
            self.assertListEqual(vs.split('GAP'), ref.split('GAP'))


class HTMLConversion(unittest.TestCase):

    a_format = '<blue><u>%s</u></blue>'

    def test_empty(self):
        R.htmlrender('')

    def test_parity(self):
        for tag in ('b', 'u', 'i'):
            markup = '<%s>stuff</%s>' % (tag, tag)
            self.assertEqual(R.html2vtml(markup), markup)

    def test_noop(self):
        self.assertEqual(R.html2vtml('<script>nope</script>'), '')

    def test_strip(self):
        self.assertEqual(R.html2vtml('<script>nope</script>'), '')
        self.assertEqual(R.html2vtml('before<script>nope</script>after'),
                         'beforeafter')

    def test_icase_tag(self):
        t = R.vtmlrender('<b>foo</b>')
        self.assertEqual(R.htmlrender('<B>foo</b>'), t)
        self.assertEqual(R.htmlrender('<B>foo</B>'), t)
        self.assertEqual(R.htmlrender('<b>foo</B>'), t)

    def test_a_tag_no_href(self):
        self.assertEqual(R.html2vtml('<a>foo</a>'), self.a_format % 'foo')

    def test_empty_href(self):
        self.assertEqual(R.html2vtml('<a href>foo</a>'), self.a_format % 'foo')

    def test_unquoted_href(self):
        self.assertEqual(R.html2vtml('<a href=link.here>foo</a>'),
                         self.a_format % 'foo (link.here)')

    def test_quoted_href(self):
        self.assertEqual(R.html2vtml('<a href="link.here">foo</a>'),
                         self.a_format % 'foo (link.here)')
        self.assertEqual(R.html2vtml("<a href='link.here'>foo</a>"),
                         self.a_format % 'foo (link.here)')

    def test_icase_href(self):
        for x in ('HREF', 'Href', 'hreF', 'href'):
            self.assertEqual(R.html2vtml('<a %s="link.here">foo</a>' % x),
                             self.a_format % 'foo (link.here)', x)
            self.assertEqual(R.html2vtml('<A %s="link.here">foo</a>' % x),
                             self.a_format % 'foo (link.here)')
            self.assertEqual(R.html2vtml('<A %s="link.here">foo</A>' % x),
                             self.a_format % 'foo (link.here)')
            self.assertEqual(R.html2vtml('<a %s="link.here">foo</A>' % x),
                             self.a_format % 'foo (link.here)')


class MDConversion(unittest.TestCase):

    def test_empty(self):
        R.mdrender('')

    def test_bold(self):
        self.assertEqual(R.mdrender('**foo**'), R.vtmlrender('\n<b>foo</b>\n'))
        self.assertEqual(R.mdrender('__foo__'), R.vtmlrender('\n<b>foo</b>\n'))



"""
Utility command for setup and exec of system completion.

Currently supports bash or zsh, but mostly bash.
"""

import argparse
import os
from .. import command


class SystemCompletion(command.Command):
    """ Generate a bash/zsh compatible completion script.

    Typically this command is run once and concatenated to your .<shell>rc
    file so completion targets for your shellish command can work from your
    system shell.  The idea is lifted directly from npm-completion. """

    name = 'completion'

    script_header = '''
        ###-begin-%(prog)s-%(name)s-###
        #
        # %(prog)s command %(name)s script
        #
        # Installation: %(prog)s %(name)s >> ~/.%(shell)src
        #
    '''

    script_body = {
        'bash': '''
            _%(prog)s_%(name)s() {
                local words cword
                cword="$COMP_CWORD"
                words=("${COMP_WORDS[@]}")
                local si="$IFS"
                IFS=$'\\n' COMPREPLY=($(COMP_CWORD="$cword" \\
                                     COMP_LINE="$COMP_LINE" \\
                                     %(prog)s %(name)s --seed "${words[@]}" \\
                                     2>/dev/null)) || return $?
                IFS="$si"
            }
            complete -o nospace -F _%(prog)s_%(name)s %(prog)s
        ''',
        'zsh': '''
            _%(prog)s_%(name)s() {
                local si=$IFS
                compadd -- $(COMP_CWORD=$((CURRENT-1)) \\
                             COMP_LINE=$BUFFER \\
                             %(prog)s %(name)s -S '' --seed "${words[@]}" \\
                             2>/dev/null)
                IFS=$si
            }
            compdef _%(prog)s_%(name)s %(name)s
        '''
    }

    script_footer = '''###-end-%(prog)s-$(name)s-###'''

    def setup_args(self, parser):
        self.add_argument('--seed', nargs=argparse.REMAINDER)
        super().setup_args(parser)

    def run(self, args):
        if not args.seed:
            return self.show_setup()
        seed = args.seed
        prog = seed.pop(0)
        index = int(os.getenv('COMP_CWORD')) - 1
        line = os.getenv('COMP_LINE')[len(prog) + 1:]
        begin = len(' '.join(seed[:index]))
        end = len(line)
        if begin > 0:
            try:
                cmd, args = self.session.cmd_split(seed[0])
            except KeyError:
                return
            cfunc = cmd.complete
        else:
            cfunc = self.session.complete_names
        for x in cfunc(seed[index], line, begin, end):
            print(x)

    def show_setup(self):
        """ Provide a helper script for the user to setup completion. """
        shell = os.getenv('SHELL')
        if not shell:
            raise SystemError("No $SHELL env var found")
        shell = os.path.basename(shell)
        if shell not in self.script_body:
            raise SystemError("Unsupported shell: %s" % shell)
        tplvars = {
            "prog": '-'.join(self.prog.split()[:-1]),
            "shell": shell,
            "name": self.name
        }
        print(self.trim(self.script_header % tplvars))
        print(self.trim(self.script_body[shell] % tplvars))
        print(self.trim(self.script_footer % tplvars))

    def trim(self, text):
        """ Trim whitespace indentation from text. """
        lines = text.splitlines()
        firstline = lines[0] or lines[1]
        indent = len(firstline) - len(firstline.lstrip())
        return '\n'.join(x[indent:] for x in lines if x.strip())




import io
import os
import readline
import shellish
import sys
import tempfile
import unittest
from shellish.command import contrib


class ArgumentCompletions(unittest.TestCase):
    """ Test completion of argument keys. """

    def completer_sig(self, line):
        """ Convert an argument line into readline function signature. """
        spliters = readline.get_completer_delims()
        for c in spliters:  # avoid regex pitfalls with brute force splitter.
            parts = line.rsplit(c, 1)
            if len(parts) == 2:
                text = parts[1]
                break
        else:
            text = line
        return (text, line, len(line), len(line))

    def complete(self, command, args):
        line = '%s %s' % (command.prog, args)
        command.get_or_create_session()
        return command.complete(*self.completer_sig(line))

    def test_empty_shows_single_opt_arg(self):
        @shellish.autocommand
        def cmd(foo=None):
            pass
        self.assertEqual(self.complete(cmd, ''), {'--foo'})

    def test_valid_shows_single_opt_arg(self):
        @shellish.autocommand
        def cmd(foo=None):
            pass
        self.assertEqual(self.complete(cmd, '-'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--f'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--fo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo'), {'--foo'})

    def test_invalid_hides_single_opt_arg(self):
        @shellish.autocommand
        def cmd(foo=None):
            pass
        self.assertEqual(self.complete(cmd, '--nope'), set())
        self.assertEqual(self.complete(cmd, '-nope'), set())
        self.assertEqual(self.complete(cmd, 'nope'), set())
        self.assertEqual(self.complete(cmd, 'Z'), set())

    def test_single_value_argument_consumed(self):
        @shellish.autocommand
        def cmd(foo=None):
            pass
        self.assertEqual(self.complete(cmd, '--foo v '), set())

    def test_one_bool_argument_consumed(self):
        cmd = shellish.Command(name='cmd')
        cmd.add_argument('--foo', action='store_true')
        self.assertEqual(self.complete(cmd, '--fo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo '), set())

    def test_many_bool_arguments_consumed(self):
        cmd = shellish.Command(name='cmd')
        cmd.add_argument('--foo', action='store_true')
        cmd.add_argument('--bar', action='store_true')
        self.assertEqual(self.complete(cmd, '--fo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo '), {'--bar'})
        self.assertEqual(self.complete(cmd, '--bar '), {'--foo'})
        self.assertEqual(self.complete(cmd, '--bar --foo '), set())

    def test_one_single_value_argument_consumed(self):
        @shellish.autocommand
        def cmd(foo=None):
            pass
        self.assertEqual(self.complete(cmd, '--fo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo value '), set())

    def test_many_single_value_arguments_consumed(self):
        @shellish.autocommand
        def cmd(foo=None, bar=None):
            pass
        self.assertEqual(self.complete(cmd, '--fo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo'), {'--foo'})
        self.assertEqual(self.complete(cmd, '--foo value '), {'--bar'})
        self.assertEqual(self.complete(cmd, '--bar value '), {'--foo'})
        self.assertEqual(self.complete(cmd, '--bar value --foo value '), set())

    def test_file_arguments(self):
        cmd = shellish.Command(name='cmd')
        cmd.add_file_argument('--foo')
        with tempfile.TemporaryDirectory() as tmp:
            files = {'./one', './two'}
            cwd = os.getcwd()
            os.chdir(tmp)
            try:
                for x in files:
                    open(x, 'w').close()
                self.assertEqual(self.complete(cmd, '--foo'), {'--foo'})
                self.assertEqual(self.complete(cmd, '--foo '), files)
                self.assertEqual(self.complete(cmd, '--foo o'), {'./one'})
                self.assertEqual(self.complete(cmd, '--foo NO'), set())
            finally:
                os.chdir(cwd)

    def test_completer_custom_parser(self):
        completer = lambda *args, **kwargs: None

        class T(shellish.Command):
            def setup_args(self, parser):
                p = parser.add_mutually_exclusive_group()
                self.add_argument('--foo', parser=p, complete=completer)
        T(name='t')


class TestSysComplete(unittest.TestCase):

    def setUp(self):
        self.stdout = sys.stdout
        sys.stdout = io.StringIO()

    def tearDown(self):
        sys.stdout = self.stdout

    def test_complete_names(self):
        abc = shellish.Command(name='abc')
        xyz = shellish.Command(name='xyz')
        root = shellish.Command(name='root')
        root.add_subcommand(abc)
        root.add_subcommand(xyz)
        sc = contrib.SystemCompletion()
        root.add_subcommand(sc)
        line = 'root a'
        os.environ["COMP_CWORD"] = '1'
        os.environ["COMP_LINE"] = line
        sc(argv='--seed %s' % line)
        self.assertEqual(sys.stdout.getvalue(), 'abc\n')

    def test_complete_command_arg(self):
        abc = shellish.Command(name='abc')
        abc.add_argument('--foo')
        root = shellish.Command(name='root')
        root.add_subcommand(abc)
        sc = contrib.SystemCompletion()
        root.add_subcommand(sc)
        line = 'root abc --f'
        os.environ["COMP_CWORD"] = '2'
        os.environ["COMP_LINE"] = line
        sc(argv='--seed %s' % line)
        self.assertEqual(sys.stdout.getvalue(), '--foo\n')



"""
The merger of argparse and cmd goes here.  This holds the main base class
used by all commands.
"""

import collections
import functools
import inspect
import itertools
import re
import shlex
from . import supplement
from .. import completer, layout, eventing, session, paging


def parse_docstring(entity):
    """ Return sanitized docstring from an entity.  The first line of the
    docstring is the title, and remaining lines are the details, aka git
    style. """
    doc = inspect.getdoc(entity)
    if not doc:
        return None, None
    doc = doc.splitlines(keepends=True)
    if not doc[0].strip():
        doc.pop(0)
    title = (doc and doc.pop(0).strip()) or None
    if doc and not doc[0].strip():
        doc.pop(0)
    desc = ''.join(doc).rstrip() or None
    return title, desc


class Command(eventing.Eventer):
    """ The primary encapsulation for a shellish command.  Each command or
    subcommand should be an instance of this class.  The docstring for sub-
    classes is used in --help output for this command. """

    name = None
    title = None
    desc = None
    use_pager = False
    ArgumentParser = supplement.ShellishParser
    ArgumentFormatter = supplement.ShellishHelpFormatter
    Session = session.Session
    completion_excludes = {'--help'}
    arg_label_fmt = '__command[%d]__'
    env_scrub_re = '[^\w\s\-_]'  # chars scrubed from env vars
    env_flatten_re = '[\s\-_]+'  # chars converted to underscores

    def setup_args(self, parser):
        """ Subclasses should provide any setup for their parsers here. """
        pass

    def prerun(self, args):
        """ Hook to do something prior to invocation. """
        pass

    def postrun(self, args, result=None, exc=None):
        """ Hook to do something following invocation. """
        pass

    def run(self, args):
        """ Primary entry point for command exec. """
        self.argparser.print_help()
        raise SystemExit(1)

    def __init__(self, parent=None, title=None, desc=None, name=None, run=None,
                 prerun=None, postrun=None, **context):
        self.add_events(['prerun', 'postrun', 'setup_args', 'precomplete',
                         'postcomplete'])
        if name:
            self.name = name
        if self.name is None:
            raise RuntimeError("Command missing `name` attribute")
        if title is not None:
            self.title = title
        if desc is not None:
            self.desc = desc
        if self.title is None and self.desc is None and \
           type(self) is not Command:
            self.title, self.desc = parse_docstring(self)
        if run is not None:
            self.run = run
        if prerun is not None:
            self.prerun = prerun
        if postrun is not None:
            self.postrun = postrun
        self.subcommands = collections.OrderedDict()
        self.default_subcommand = None
        self.session = None
        self.context_keys = set()
        self._autoenv_actions = set()
        self.inject_context(context)
        self.subparsers = None
        self.argparser = self.create_argparser()
        self.parent = parent
        self.setup_args(self.argparser)
        self.fire_event('setup_args', self.argparser)

    def get_or_create_session(self):
        if self.session is None:
            self.attach_session()
        return self.session

    def parse_args(self, argv=None):
        """ Return an argparse.Namespace of the argv string or sys.argv if
        argv is None. """
        arg_input = shlex.split(argv) if argv is not None else None
        self.get_or_create_session()
        return self.argparser.parse_args(arg_input)

    def __call__(self, args=None, argv=None):
        """ If a subparser is present and configured  we forward invocation to
        the correct subcommand method. If all else fails we call run(). """
        session = self.get_or_create_session()
        if args is None:
            args = self.parse_args(argv)
        commands = self.get_commands_from(args)
        if self.subparsers:
            try:
                command = commands[self.depth]
            except IndexError:
                pass
            else:
                return command(args)
            if self.default_subcommand:
                parser = self.default_subcommand.argparser
                parser.parse_args([], namespace=args)
                return self(args)  # retry
        return session.execute(self, args)

    def __getitem__(self, item):
        return self.subcommands[item]

    def get_pager_spec(self):
        """ Find the best pager settings for this command.  If the user has
        specified overrides in the INI config file we prefer those. """
        self_config = self.get_config()
        pagercmd = self_config.get('pager')
        istty = self_config.getboolean('pager_istty')
        core_config = self.get_config('core')
        if pagercmd is None:
            pagercmd = core_config.get('pager')
        if istty is None:
            istty = core_config.get('pager_istty')
        return {
            "pagercmd": pagercmd,
            "istty": istty
        }

    def run_wrap(self, args):
        """ Wrap some standard protocol around a command's run method.  This
        wrapper should generally never capture exceptions.  It can look at
        them and do things but prerun and postrun should always be symmetric.
        Any exception suppression should happen in the `session.execute`. """
        self.fire_event('prerun', args)
        self.prerun(args)
        try:
            if self.session.allow_pager and self.use_pager:
                desc = 'Command\: %s' % '-'.join(self.prog.split())
                with paging.pager_redirect(desc, **self.get_pager_spec()):
                    result = self.run(args)
            else:
                result = self.run(args)
        except (SystemExit, Exception) as e:
            self.postrun(args, exc=e)
            self.fire_event('postrun', args, exc=e)
            raise e
        else:
            self.postrun(args, result=result)
            self.fire_event('postrun', args, result=result)
            return result

    def config_section(self):
        """ The string used in a .<ROOT>_config file section.  Usually this
        is just the full prog name for a command minus the root command. """
        names = []
        cmd = self
        while cmd.parent:
            names.append(cmd.name)
            cmd = cmd.parent
        if not names:
            return self.name
        else:
            return ' '.join(reversed(names))

    def default_config(self):
        """ Can be overridden to provide a 1 level deep dictionary of config
        values.  Theses values are optionally overridden by the end-user via
        the session's load_config routine, that essentially looks for an INI
        file where the `[section]` is the `.prog` value for this command. """
        return {}

    def get_config(self, section=None):
        """ Return the merged end-user configuration for this command or a
        specific section if set in `section`. """
        config = self.session.config
        section = self.config_section() if section is None else section
        try:
            return config[section]
        except KeyError:
            config.add_section(section)
            return config[section]

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        """ Copy context from the parent into this instance as well as
        adjusting or depth value to indicate where we exist in a command
        tree. """
        self._parent = parent
        if parent:
            pctx = dict((x, getattr(parent, x)) for x in parent.context_keys)
            self.inject_context(pctx)
            self.depth = parent.depth + 1
            for command in self.subcommands.values():
                command.parent = self  # bump.
        else:
            self.depth = 0

    def find_root(self):
        """ Traverse parent refs to top. """
        cmd = self
        while cmd.parent:
            cmd = cmd.parent
        return cmd

    def inject_context(self, __context_dict__=None, **context):
        """ Map context dict to this instance as attributes and keep note of
        the keys being set so we can pass this along to any subcommands. """
        context = context or __context_dict__
        self.context_keys |= set(context.keys())
        for key, value in context.items():
            setattr(self, key, value)
        for command in self.subcommands.values():
            command.inject_context(context)

    @property
    def prog(self):
        return self.argparser.prog

    @prog.setter
    def prog(self, prog):
        """ Update ourself and any of our subcommands. """
        self.argparser.prog = prog
        fmt = '%s %%s' % prog if prog else '%s'
        for command in self.subcommands.values():
            command.prog = fmt % command.name
        # Rebind autoenv vars with recomputed key.
        for action in self._autoenv_actions:
            self.argparser.unbind_env(action)
            self.argparser.bind_env(action, self._make_autoenv(action))

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, value):
        """ Update ourself and any of our subcommands. """
        for command in self.subcommands.values():
            command.depth = value + 1
            del command.argparser._defaults[self.arg_label_fmt % self._depth]
            command.argparser._defaults[self.arg_label_fmt % value] = command
        self._depth = value

    def add_argument(self, *args, parser=None, autoenv=False, env=None,
                     complete=None, **kwargs):
        """ Allow cleaner action supplementation.  Autoenv will generate an
        environment variable to be usable as a defaults setter based on the
        command name and the dest property of the action. """
        if parser is None:
            parser = self.argparser
        action = parser.add_argument(*args, **kwargs)
        if autoenv:
            if env is not None:
                raise TypeError('Arguments `env` and `autoenv` are mutually '
                                'exclusive')
            env = self._make_autoenv(action)
        if env:
            self.argparser.bind_env(action, env)
            if autoenv:
                self._autoenv_actions.add(action)
        if complete:
            action.complete = complete
        return action

    def _make_autoenv(self, action):
        """ Generate a suitable env variable for this action.  This is
        dependant on our subcommand hierarchy.  Review the prog setter for
        details. """
        env = ('%s_%s' % (self.prog, action.dest)).upper()
        env = re.sub(self.env_scrub_re, '', env.strip())
        env = re.sub(self.env_flatten_re, '_', env)
        if re.match('^[0-9]', env):
            # Handle leading numbers.
            env = '_%s' % env
        return env

    def add_file_argument(self, *args, mode='r', buffering=1,
                          filetype_options=None, **kwargs):
        """ Add a tab-completion safe FileType argument.  This argument
        differs from a normal argparse.FileType based argument in that the
        value is a factory function that returns a file handle instead of
        providing an already open file handle.  There are various reasons
        why this is a better approach but it is also required to avoid
        erroneous creation of files with shellish tab completion. """
        type_ = supplement.SafeFileType(mode=mode, bufsize=buffering,
                                        **filetype_options or {})
        return self.add_argument(*args, type=type_, **kwargs)

    def add_table_arguments(self, *args, parser=None, **kwargs):
        if parser is None:
            parser = self.argparser
        return layout.Table.attach_arguments(parser, *args, **kwargs)

    def create_argparser(self):
        """ Factory for arg parser.  Can be overridden as long as it returns
        an ArgParser compatible instance. """
        if self.desc:
            if self.title:
                fulldesc = '%s\n\n%s' % (self.title, self.desc)
            else:
                fulldesc = self.desc
        else:
            fulldesc = self.title
        return self.ArgumentParser(command=self, prog=self.name,
                                   description=fulldesc)

    def attach_session(self):
        """ Create a session and inject it as context for this command and any
        subcommands. """
        assert self.session is None
        root = self.find_root()
        session = self.Session(root)
        root.inject_context(session=session)
        return session

    def complete(self, text, line, begin, end):
        """ Get and format completer choices.  Note that all internal calls to
        completer functions must use [frozen]set()s. """
        self.fire_event('precomplete', text, line, begin, end)
        choices = self._complete(text, line, begin, end)
        choices -= self.completion_excludes
        sz = len(choices)
        if sz == 1:
            # XXX: This is pretty bad logic here.  In reality we need much
            # more complicated escape handling and imbalanced quote support.
            return set(shlex.quote(x) for x in choices)
        elif sz > 2:
            # We don't need the sentinel choice to prevent completion
            # when there is already more than 1 choice.
            choices -= {completer.ActionCompleter.sentinel}
        self.fire_event('postcomplete', choices)
        return choices

    def _complete(self, text, line, begin, end):
        """ Do naive argument parsing so the completer has better ability to
        understand expansion rules. """
        line = line[:end]  # Ignore characters following the cursor.
        fullargs = self.split_line(line)[1:]
        args = fullargs[:]
        options = self.deep_scan_parser(self.argparser)

        # Walk into options tree if subcommands are detected.
        last_subcommand = None
        while True:
            for key, completers in options.items():
                if key in args and hasattr(completers[0], 'items'):
                    args.remove(key)
                    last_subcommand = key
                    options = completers[0]
                    break
            else:
                break

        if text == last_subcommand:
            # We have to specially catch the case where the last argument is
            # the key used to find our subparser.  More specifically when the
            # cursor is not preceded by a space too, as this prevents the
            # completion routines from continuing.  The simplest way without
            # complicating the algo for coming up with our options list is to
            # simply shortcut the completer by returning a single item.
            # Subsequent tabs will work normally.
            return {text}

        # Look for incomplete actions.
        choices = set(x for x in options
                      if x is not None and x.startswith(text))
        arg_buf = []
        pos_args = []
        trailing_action = None
        # The slice below skips the last arg if it is 'active'.
        for x in reversed(args[:-1 if text else None]):
            if x in options:
                action = options[x][0]
                action.consume(arg_buf)
                pos_args.extend(arg_buf)
                del arg_buf[:]
                if action.full:
                    choices -= {action.key}
                if not trailing_action:
                    trailing_action = action
                    if not action.full:
                        if action.reached_min:
                            choices |= action(self, text, fullargs)
                            choices -= {action.key}
                        else:
                            choices = action(self, text, fullargs)
                            break
            else:
                arg_buf.insert(0, x)
        pos_args.extend(arg_buf)

        # Feed any remaining arguments in the buffer to positionals so long as
        # there isn't a trailing action that can still consume.
        if None in options and (not trailing_action or trailing_action.full):
            for x_action in options[None]:
                x_action.consume(pos_args)
                if not x_action.reached_min:
                    choices = x_action(self, text, fullargs)
                    break
                elif not x_action.full:
                    choices |= x_action(self, text, fullargs)
        return choices

    def split_line(self, line):
        """ Try to do pure shlex.split unless it can't parse the line. In that
        case we trim the input line until shlex can split the args and tack the
        unparsable portion on as the last argument. """
        remainder = []
        while True:
            try:
                args = shlex.split(line)
            except ValueError:
                remainder.append(line[-1])
                line = line[:-1]
            else:
                if remainder:
                    args.append(''.join(reversed(remainder)))
                return args

    @functools.lru_cache()
    def deep_scan_parser(self, parser):
        results = collections.defaultdict(list)
        for x in parser._actions:
            ac = completer.ActionCompleter(x)
            if ac.subparsers:
                for key, xx in ac.subparsers.items():
                    results[key].append(self.deep_scan_parser(xx))
            else:
                results[ac.key].append(ac)
        return results

    def get_commands_from(self, args):
        """ We have to code the key names for each depth.  This method scans
        for each level and returns a list of the command arguments. """
        commands = []
        for i in itertools.count(0):
            try:
                commands.append(getattr(args, self.arg_label_fmt % i))
            except AttributeError:
                break
        return commands

    def add_subcommand(self, command, default=False):
        if isinstance(command, type):
            command = command()
        command.parent = self
        if command.name is None:
            raise TypeError('Cannot add unnamed command: %s' % command)
        if command.name in self.subcommands:
            raise ValueError('Command name already added: %s' % command.name)
        if not self.subparsers:
            desc = 'Provide a subcommand argument to perform an operation.'
            addsub = self.argparser.add_subparsers
            self.subparsers = addsub(title='subcommands', description=desc,
                                     metavar='COMMAND')
        if default:
            if self.default_subcommand:
                raise ValueError("Default subcommand already exists.")
            self.default_subcommand = command
        if command.title is not None:
            help_fmt = '%s (default)' if default else '%s'
            help = help_fmt % command.title
        else:
            help = '(default)' if default else ''
        command.prog = '%s %s' % (self.prog, command.name)
        command.argparser._defaults[self.arg_label_fmt % self.depth] = command
        action = self.subparsers._ChoicesPseudoAction(command.name, (), help)
        self.subparsers._choices_actions.append(action)
        self.subparsers._name_parser_map[command.name] = command.argparser
        self.subcommands[command.name] = command

    def remove_subcommand(self, command=None, name=None):
        if name is None:
            if command is None:
                raise TypeError('A command or name is required')
            name = command.name
        command = self.subcommands.pop(name)
        del self.subparsers._name_parser_map[name]
        for action in self.subparsers._choices_actions:
            if action.dest == name:
                break
        else:
            raise RuntimeError("Subparser action not found for subcommand")
        self.subparsers._choices_actions.remove(action)
        command.session = None
        command.parent = None



#!/usr/bin/env python
from setuptools import setup, find_packages


def long_desc():
    with open('README.md') as f:
        return f.read()


setup(
    name='shellish',
    version='5.1',
    description='A framework for CLI/shell programs.',
    author='Justin Mayfield',
    author_email='tooker@gmail.com',
    url='https://github.com/mayfield/shellish/',
    license='MIT',
    long_description=long_desc(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=['markdown2'],
    test_suite='test',
    entry_points={
        'console_scripts': [
            'csvpretty=shellish.tools.csvpretty:csvpretty',
            'mdcat=shellish.tools.mdcat:mdcat'
        ]
    },
    classifiers=[
        'Development Status :: 6 - Mature',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Shells',
    ]
)



"""
Sanity tests for the traceback formatting
"""

import io
import shellish
import unittest


class TracebackSanity(unittest.TestCase):

    def assertTracebackFormat(self, traceback_lines, exc_name):
        """ Rougly assert that a list of traceback lines looks okay. """
        self.assertTrue(traceback_lines)
        self.assertIn('Traceback (most recent call last)', traceback_lines[0])
        self.assertIn('File', traceback_lines[1])
        self.assertIn('line', traceback_lines[1])
        self.assertIn('1.', traceback_lines[-3])
        self.assertIn('File', traceback_lines[-3])
        self.assertIn('line', traceback_lines[-3])
        self.assertIn(exc_name, traceback_lines[-1])

    def stack_push(self, levels, callback, *args, **kwargs):
        if levels:
            return self.stack_push(levels - 1, callback, *args, **kwargs)
        else:
            return callback(*args, **kwargs)

    def test_format_single(self):
        def raiser():
            raise ValueError('foo')
        try:
            self.stack_push(10, raiser)
        except ValueError as e:
            self.assertTracebackFormat(list(shellish.format_exception(e)),
                                       'ValueError')

    def test_format_1stack(self):
        def raiser():
            raise ValueError('foo')
        try:
            self.stack_push(1, raiser)
        except ValueError as e:
            self.assertTracebackFormat(list(shellish.format_exception(e)),
                                       'ValueError')

    def test_format_10stack(self):
        def raiser():
            raise ValueError('foo')
        try:
            self.stack_push(10, raiser)
        except ValueError as e:
            self.assertTracebackFormat(list(shellish.format_exception(e)),
                                       'ValueError')

    def test_format_with_cause(self):
        try:
            raise ValueError('foo')
        except ValueError as outer:
            try:
                raise TypeError('bar')
            except TypeError as inner:
                self.assertTracebackFormat(
                    list(shellish.format_exception(inner)), 'TypeError')

    def test_print(self):
        buf = io.StringIO()
        try:
            raise ValueError('foo')
        except ValueError as e:
            shellish.print_exception(e, file=buf)
            self.assertTrue(buf.getvalue())



from .completer import *
from .command import *
from .layout import *
from .data import *
from .eventing import *
from .session import *
from .paging import *
from .rendering import *

from . import logging



import shellish
import unittest
import tempfile


class TTYPagingTests(unittest.TestCase):

    def test_small(self):
        with tempfile.TemporaryDirectory() as tdir:
            outfile = '%s/out' % tdir
            with shellish.pager_redirect('test', pagercmd='head -n1 > %s' %
                                         outfile):
                for x in range(10):
                    try:
                        print(x)
                    except BrokenPipeError:
                        break
            with open(outfile) as f:
                self.assertEqual(f.read(), '0\n')

    def test_pipe_overflow(self):
        with tempfile.TemporaryDirectory() as tdir:
            outfile = '%s/out' % tdir
            pipe_break = False
            with shellish.pager_redirect('test', pagercmd='head -n1 > %s' %
                                         outfile):
                for x in range(10000):
                    try:
                        print(('%d' % x) * 1000)
                    except BrokenPipeError:
                        pipe_break = True
                        break
            self.assertTrue(pipe_break)
            with open(outfile) as f:
                self.assertEqual(f.read().rstrip(), '0' * 1000)

    def test_nested(self):
        with tempfile.TemporaryDirectory() as tdir:
            outfile1 = '%s/out1' % tdir
            outfile2 = '%s/out2' % tdir
            with shellish.pager_redirect('test', pagercmd='cat > %s' %
                                         outfile1):
                print('outer')
                with shellish.pager_redirect('test', pagercmd='cat > %s' %
                                             outfile2):
                    print('inner')
            self.assertRaises(IOError, open, outfile2)
            with open(outfile1) as f:
                self.assertEqual(list(f), ['outer\n', 'inner\n'])
